/*-
 * Copyright (c) 1993, 1994 Michael B. Durian.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Michael B. Durian.
 * 4. The name of the the Author may be used to endorse or promote 
 *    products derived from this software without specific prior written 
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "midi.h"
#if NMIDI > 0
#include "param.h"
#include "systm.h"
#include "ioctl.h"
#include "tty.h"
#include "proc.h"
#include "user.h"
#include "file.h"
#include "uio.h"
#include "malloc.h"
#include "device.h"
#include "syslog.h"
#include "kernel.h"
#include "vnode.h"

#ifndef __bsdi__
#include "i386/isa/isa_device.h"
#include <machine/pio.h>
#endif

#include "midivar.h"
#include "midiioctl.h"
#include "isavar.h"
#include "icu.h"

/* I've been burned too many times by different byte ordering */
#define OUTB(x, y) outb((x), (y))
#define INB(x) inb((x))

struct midi_softc {      /* Driver status information */
#ifdef __bsdi__
        struct  device sc_dev;  /* base device */
        struct  isadev sc_id;   /* ISA device */
        struct  intrhand sc_ih; /* interrupt handler */
#endif

	/* midi specific stuff */
	struct	selinfo wsel;
	struct	selinfo rsel;
	struct	event_queue *rqueue;
	struct	event_queue *wqueue;
	struct	stynamic rpartial;
	struct	stynamic wpartial;
	struct	partial_event partial_event;
	struct	timeval prev_incoming;
	volatile	long status;
	long long	premainder;
	long long	rremainder;
	long	ptempo;			/* usec / beat */
	long	rtempo;			/* usec / beat */
	long	prev_rtempo;
	int	pgid;
	int	addr;
	int	division;
	int	wpartialpos;
	long	write_noop_time;
	short	noteon[0x80];		/*
					 * each element is a different pitch
					 * each bit is a different channel
					 */
	u_char	readrs;
	u_char	writers;
	u_char	noteonrs;
#ifdef __bsdi__
};
#else
} midi_sc[NMIDI];
#endif

#ifdef __bsdi__
int midiprobe __P((struct device *parent, struct cfdata *cf, void *aux));
void midiattach __P((struct device *parent, struct device *dev, void *args));
#else
int midiprobe __P((struct isa_device *dev));
int midiattach __P((struct isa_device *dev));
#endif
int midiopen __P((dev_t dev, int flags, int fmt, struct proc *p));
int midiclose __P((dev_t dev, int flags, int fmt, struct proc *p));
int midiread __P((dev_t dev, struct uio *u, int flag));
int midiioctl __P((dev_t dev, int cmd, caddr_t arg, int mode, struct proc *p));
int midiselect __P((dev_t dev, int rw, struct proc *p));
#ifdef __bsdi__
int midiintr __P((struct midi_softc *softc));
static void midiforceintr __P((struct midi_softc *));
#else
int midiintr __P((int unit));
#endif
static void midi_copy_event __P((struct event *, struct event *));
static void midi_initq __P((struct event_queue *));
static int midi_deq __P((struct event_queue *, struct event **));
static int midi_peekq __P((struct event_queue *, struct event **));
static int midi_enq __P((struct event_queue *, struct event *));
static int midi_reset __P((struct midi_softc *));
static int midi_wait_rdy_rcv __P((struct midi_softc *softc));
static int midi_send_command __P((struct midi_softc *softc,
    /* u_char */ int comm));
static int midi_uart __P((struct midi_softc *));
static int midi_event2smf __P((struct midi_softc *, struct event *,
    struct stynamic *));
static int midi_next_byte __P((struct midi_softc *, struct uio *));
static int midi_uio2event __P((struct midi_softc *, struct uio *,
    struct event *));
static int midi_fix2var __P((u_long, u_char *));
static int midi_var2fix __P((u_char *, u_long *));
static u_long midi_smf2kernel_tick __P((struct midi_softc *, long));
static u_long midi_kernel2smf_tick __P((struct midi_softc *, long, long));
static void midi_timeout __P((caddr_t arg));
static void midi_add_complete_event __P((struct midi_softc *));
static void midi_schedule_timeout __P((struct midi_softc *));
static void midi_write_event __P((struct midi_softc *, struct event *));
static void midi_reset_devices __P((struct midi_softc *));
static void stynamic_add_byte __P((struct stynamic *, int /* u_char */));
static void stynamic_add_bytes __P((struct stynamic *, u_char *, int));
static u_char stynamic_get_byte __P((struct stynamic *, int));
static void stynamic_copy __P((struct stynamic *, void *, int));
static void stynamic_copy_from __P((struct stynamic *, int, void *, int));
static void stynamic_append __P((struct stynamic *, struct stynamic *));
static void stynamic_release __P((struct stynamic *));
static void stynamic_shift __P((struct stynamic *, int));
static void stynamic_init __P((struct stynamic *));
static void stynamic_print __P((struct stynamic *));
static int midi_killable __P((struct proc *owner, int pid));

#ifdef __bsdi__
struct cfdriver midicd =
        { NULL, "midi", midiprobe, midiattach, sizeof(struct midi_softc) };
#else
struct isa_driver mididriver = {midiprobe, midiattach, "midi"};
#endif

#ifdef __bsdi__
int
midiprobe(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
#else
int
midiprobe(dvp)
	struct isa_device *dvp;
#endif
{
#ifdef __bsdi__
	register struct isa_attach_args *ia = (struct isa_attach_args *)aux;
#endif
	struct midi_softc softc;

	/*
	 * We're faking a temp softc here.  All it really needs is
	 * the addr field, but we supply it so midiforceintr can
	 * call midi_reset
	 */
#ifdef __bsdi__
	softc.addr = ia->ia_iobase;
	/* should be able to do a reset if it exists */
	if (ia->ia_irq == IRQUNK) {
		ia->ia_irq = isa_discoverintr(midiforceintr, &softc);
		if (ffs(ia->ia_irq) - 1 == 0)
			return (0);
	}
#else
	softc.addr = dvp->id_iobase;
#endif

	if (!midi_reset(&softc))
		return (0);

#ifdef __bsdi__
	/* We only use addr and addr + 1 */
	ia->ia_iosize = 2;
#endif

	return (1);
}

#ifdef __bsdi__
void
midiforceintr(softc)
	struct midi_softc *softc;
{

	/*
	 * If the first reset fails, it's because we are in UART
	 * mode.  UART mode does not generate interrupts or
	 * acks for commands.  Thus the first reset takes the
	 * board out of UART mode and the second will cause
	 * an ack and an interrupt.
	 */
	if (!midi_reset(softc))
		midi_reset(softc);
}
#endif

#ifdef __bsdi__
void
midiattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
#else
int
midiattach(isdp)
	struct isa_device *isdp;
#endif
{ 
#ifdef __bsdi__
	register struct isa_attach_args *ia = (struct isa_attach_args *)aux;
	register struct midi_softc *softc = (struct midi_softc *)self;
#else
	register struct midi_softc *softc;

	softc = &midi_sc[isdp->id_unit];
#endif

	softc->status = MIDI_RD_BLOCK;
#ifdef __bsdi__
	softc->addr = ia->ia_iobase;
#else
	softc->addr = isdp->id_iobase;
#endif
	printf("\n");
	/*
	 * The board goes into UART mode, forever to stay.
	 */
	if (!midi_uart(softc)) {
		log(LOG_ERR, "Couldn't put MPU401 into UART mode!\n");
		softc->status |= MIDI_UNAVAILABLE;
	}

	/* allocate memory for event queues */
	if ((softc->rqueue = malloc(sizeof(struct event_queue), M_DEVBUF,
	    M_NOWAIT)) == NULL) {
		log(LOG_ERR, "No memory for rqueue\n");
		softc->status |= MIDI_UNAVAILABLE;
	}
	if ((softc->wqueue = malloc(sizeof(struct event_queue), M_DEVBUF,
	    M_NOWAIT)) == NULL) {
		log(LOG_ERR, "No memory for wqueue\n");
		softc->status |= MIDI_UNAVAILABLE;
		free(softc->rqueue, M_DEVBUF);
	}
	/* zero read/write queue to clear stynamic structures */
	bzero(softc->rqueue, sizeof(struct event_queue));
	bzero(softc->wqueue, sizeof(struct event_queue));
	stynamic_init(&softc->rpartial);
	stynamic_init(&softc->wpartial);
	softc->wpartialpos = 0;

#ifdef __bsdi__
	isa_establish(&softc->sc_id, &softc->sc_dev);
        softc->sc_ih.ih_fun = midiintr;
        softc->sc_ih.ih_arg = softc;
	intr_establish(ia->ia_irq, &softc->sc_ih, DV_TTY);
#else
	return (1);
#endif
}

int
midiopen(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	register struct midi_softc *softc;
	register int unit;
 
	unit = minor(dev);
#ifdef __bsdi__
	if (unit >= midicd.cd_ndevs || (softc = midicd.cd_devs[unit]) == NULL)
		return (ENXIO);
#else
	if (unit > NMIDI)
		return (ENXIO);
	softc = &midi_sc[unit];
#endif
	if (softc->status & MIDI_UNAVAILABLE)
		return (EIO);
	
	if (softc->status & MIDI_OPEN)
		return (EBUSY);
	else
		softc->status = MIDI_OPEN;

	/*
	 * reads will block until something appears
	 */
	softc->status |= MIDI_RD_BLOCK;

	/* initialize the queues */
	midi_initq(softc->rqueue);
	midi_initq(softc->wqueue);

	softc->partial_event.event.len = 0;
	softc->partial_event.event.datad = NULL;

	/* make sure we are in UART mode */
	if (!midi_fullreset(softc)) {
		log(LOG_ERR, "Couldn't put MPU401 into UART mode!\n");
		softc->status |= MIDI_UNAVAILABLE;
		free(softc->rqueue, M_DEVBUF);
		free(softc->wqueue, M_DEVBUF);
		return (EIO);
	}

	softc->pgid = p->p_pid;

	/* are we going to read, write or both? */
	if (flag & FREAD)
		softc->status |= MIDI_READING;
	else
		softc->status &= ~MIDI_READING;
	if (flag & FWRITE)
		softc->status |= MIDI_WRITING;
	else
		softc->status &= ~MIDI_WRITING;

	return (0);
}

int
midiclose(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	register struct midi_softc *softc;
	register int error, unit;

	unit = minor(dev);
#ifdef __bsdi__
	softc = midicd.cd_devs[unit];
#else
	softc = &midi_sc[unit];
#endif

	/*
	 * we're not going to finish closing until everything has
	 * been played.
	 */
	error = 0;
	if (softc->status & MIDI_WRITING) {
		/* same as MFLUSHQ ioctl */
		if (softc->wqueue->count != 0) {
			softc->status |= MIDI_FLUSH_SLEEP;
			do {
				if (error = tsleep(&softc->status,
				    PWAIT | PCATCH, "midiclose", 0))
					softc->status &= ~MIDI_FLUSH_SLEEP;
			} while (softc->status & MIDI_FLUSH_SLEEP);
		}
	}

	/* turn off any notes that might be stuck on */
	midi_reset_devices(softc);
	midi_fullreset(softc);

	softc->status &= ~MIDI_OPEN;
	return (error);
}

int
midiread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct event *event;
	register struct midi_softc *softc;
	int error, in, num_to_move, unit;

	unit = minor(dev);
#ifdef __bsdi__
	softc = midicd.cd_devs[unit];
#else
	softc = &midi_sc[unit];
#endif

	if (softc->rqueue->count == 0 && softc->rpartial.len == 0) {
		if (flag & IO_NDELAY)
			return (EWOULDBLOCK);
		else {
			softc->status |= MIDI_RD_SLEEP;
			do {
				if (error = tsleep(softc->rqueue,
				    PWAIT | PCATCH, "midiread", 0))
					return (error);
			} while (softc->status & MIDI_RD_SLEEP);
			/* XXX maybe check for abort here */
		}
	}
	while (uio->uio_resid) {
		/*
		 * dequeue an event if partial is empty
		 */
		if (softc->rpartial.len == 0) {
			if (!midi_deq(softc->rqueue, &event)) {
				softc->status |= MIDI_RD_BLOCK;
				return (0);
			}
			midi_event2smf(softc, event, &softc->rpartial);
			stynamic_release(&event->data);
		}
		/* read out as much of rpartial as possible */
		num_to_move = MIN(softc->rpartial.len, uio->uio_resid);
		if (softc->rpartial.len <= STYNAMIC_SIZE) {
			if ((error = uiomove(softc->rpartial.datas,
			    num_to_move, uio)) == -1)
				return (error);
		} else {
			if ((error = uiomove(softc->rpartial.datad,
			    num_to_move, uio)) == -1)
				return (error);
		}
		stynamic_shift(&softc->rpartial, num_to_move);
	}
	return (0);
}

int
midiwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	register struct midi_softc *softc;
	struct event event;
	int convert, error, unit, iovec, empty_queue;

	unit = minor(dev);
#ifdef __bsdi__
	softc = midicd.cd_devs[unit];
#else
	softc = &midi_sc[unit];
#endif

	if (softc->wqueue->count == 0)
		empty_queue = 1;
	else
		empty_queue = 0;

	/* check to see if we'll block */
	if (softc->status & MIDI_WR_BLOCK) {
		if (flag & IO_NDELAY)
			return (EWOULDBLOCK);
		else {
			softc->status |= MIDI_WR_SLEEP;
			do {
				if (error = tsleep(softc->wqueue,
				    PWAIT | PCATCH, "midiwrite", 0))
					return (error);
			} while (softc->status & MIDI_WR_SLEEP);
		}
	}
	/* if returns from sleep and should abort because of DRAIN */
	if (softc->status & MIDI_WR_ABORT) {
		softc->status &= ~MIDI_WR_ABORT;
		return (0);
	}

	stynamic_init(&event.data);
	while (uio->uio_resid) {
		if (softc->wqueue->count == MIDI_Q_SIZE) {
			softc->status |= MIDI_WR_BLOCK;
			if (empty_queue) {
				midi_schedule_timeout(softc);
				empty_queue = 0;
			}
			if (flag & IO_NDELAY)
				return (0);
			else {
				softc->status |= MIDI_WR_SLEEP;
				do {
					if (error = tsleep(softc->wqueue,
					    PWAIT | PCATCH, "midiwrite", 0))
						return (error);
				} while (softc->status & MIDI_WR_SLEEP);
			}
		}

		/*
		 * 1) get a complete event off queue
		 * 2) convert it from SMF to board format
		 * 3) deal with it
		 */
		convert = midi_uio2event(softc, uio, &event);
		switch (convert) {
		case 0:
			break;
		case -1:
			/* not a complete event - we'll get it next time */
			if (empty_queue)
				midi_schedule_timeout(softc);
			return (0);
		case -2:
			/* it was a noop event */
			break;
		default:
			return (convert);
		}
		if (convert == -2)
			continue;

		if (midi_enq(softc->wqueue, &event) == -1)
			return (EIO);
		stynamic_release(&event.data);
		/* set flag so next time we cross LOW water we will SIGIO */
		if (softc->wqueue->count >= MIDI_LOW_WATER)
			softc->status |= MIDI_SENDIO;
	}

	if (empty_queue)
		midi_schedule_timeout(softc);

	return (0);
}

#ifdef __bsdi__
int
midiintr(softc)
	struct midi_softc *softc;
#else
int
midiintr(unit)
	int unit;
#endif
{
#ifndef __bsdi__
	register struct midi_softc *softc;
#endif
	register u_char code;
	register struct partial_event *pe;

#ifndef __bsdi__
	softc = &midi_sc[unit];
#endif

	code = INB(softc->addr + MIDI_DATA);
	if (code == 0xfe && (softc->status & MIDI_NEED_ACK)) {
		softc->status &= ~MIDI_NEED_ACK;
		return (1);
	}
	/* throw away data if no one has the device open */
	if (!(softc->status & MIDI_READING))
		return (1);

	/* pass input data to out port if necessary */
	if (softc->status & MIDI_THRU) {
		if (!midi_wait_rdy_rcv(softc))
			log(LOG_ERR, "Couldn't pass data thru\n");
		else
			OUTB(softc->addr + MIDI_DATA, code);
	}
	pe = &softc->partial_event;

	/* check for realtime events */
	if ((code & 0xf8) == 0xf8) {
		switch (code) {
		case 0xfa:	/* start */
			/* reset prev_incoming */
			softc->prev_incoming.tv_sec = time.tv_sec;
			softc->prev_incoming.tv_usec = time.tv_usec;
			break;
		case 0xff:	/* reset */
			/*
			 * I don't really want to do a full reset
			 * I'll just clear any event in progress
			 */
			stynamic_release(&pe->event);
			pe->state = START;
			break;
		case 0xf8:	/* timing clock */
		case 0xf9:	/* undefined */
		case 0xfb:	/* continue */
		case 0xfc:	/* stop */
		case 0xfd:	/* undefined */
		case 0xfe:	/* active sensing */
			break;
		}
		return (1);
	}
INTR_SWITCH:
	switch (pe->state) {
	case START:
		/* record when the time when the event started */
		pe->time.tv_sec = time.tv_sec;
		pe->time.tv_usec = time.tv_usec;
		pe->tempo = softc->rtempo;
		/* start collecting the input */
		stynamic_release(&pe->event);
		/* decide what state is next */
		if (!(code & 0x80)) {
			stynamic_add_byte(&pe->event, code);
			switch (pe->rs) {
			case 0x80:
			case 0x90:
			case 0xa0:
			case 0xb0:
			case 0xe0:
				/*
				 * code is the first data byte, but
				 * we still need to get the second
				 */
				pe->state = NEEDDATA2;
				break;
			case 0xc0:
			case 0xd0:
				/* code is the only data byte */
				pe->state = START;
				midi_add_complete_event(softc);
				break;
			default:
				break;
			}
		} else {
			switch (code & 0xf0) {
			case 0x80:
			case 0x90:
			case 0xa0:
			case 0xb0:
			case 0xe0:
				pe->rs = code & 0xf0;
				stynamic_add_byte(&pe->event, code);
				pe->state = NEEDDATA1;
				break;
			case 0xc0:
			case 0xd0:
				pe->rs = code & 0xf0;
				stynamic_add_byte(&pe->event, code);
				pe->state = NEEDDATA2;
				break;
			default:
				switch (code) {
				case 0xf0: /* sysex */
					stynamic_add_byte(&pe->event, code);
					pe->state = SYSEX;
					break;
				case 0xf1: /* undefined */
				case 0xf4: /* undefined */
				case 0xf5: /* undefined */
				case 0xf6: /* tune request */
				case 0xf7: /* EOX (terminator) */
					/* ignore these */
					break;
				case 0xf2: /* song position */
					pe->state = SYSTEM2;
					break;
				case 0xf3: /* song select */
					pe->state = SYSTEM1;
					break;
				}
				break;
			}
		}
		break;
	case NEEDDATA1:
		stynamic_add_byte(&pe->event, code);
		pe->state = NEEDDATA2;
		break;
	case NEEDDATA2:
		stynamic_add_byte(&pe->event, code);
		pe->state = START;
		midi_add_complete_event(softc);
		break;
	case SYSEX:
		/* any non-data byte ends sysex */
		if (!(code & 0x80))
			stynamic_add_byte(&pe->event, code);
		else {
			stynamic_add_byte(&pe->event, 0xf7);
			midi_add_complete_event(softc);
			pe->state = START;
			if (code != 0xf7)
				goto INTR_SWITCH;
		}
		break;
	case SYSTEM1:
		/* throw away one data byte of a system message */
		pe->state = START;
		break;
	case SYSTEM2:
		/* throw away two data bytes of a system message */
		pe->state = SYSTEM1;
		break;
	}
	return (1);
}

int
midiioctl(dev, cmd, data, flag, p)
	dev_t dev;
	int cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	struct event *event;
	struct midi_softc *softc;
	register int error, unit;
	int i, ret, val;

	unit = minor(dev);
#ifdef __bsdi__
	softc = midicd.cd_devs[unit];
#else
	softc = &midi_sc[unit];
#endif

	switch (cmd) {
	case FIOASYNC:
	case MASYNC:
		/*
		 * Linux doesn't properly process the FIOASYNC
		 * ioctl entry point, thus we have two.  For
		 * compatibility purposes, I'm duplicating it
		 * here too.
		 */
		if (!*(int *)data)
			softc->status &= ~MIDI_ASYNC;
		else {
			struct proc *p;

			softc->status |= MIDI_ASYNC;
			if (softc->wqueue->count < MIDI_LOW_WATER
			    && softc->status & MIDI_WRITING) {
				if (softc->pgid < 0)
					gsignal(-softc->pgid, SIGIO);
				else if ((p = pfind(softc->pgid)) != 0)
					psignal(p, SIGIO);
			}
		}
		break;
	case TIOCSPGRP:
		val = *(int *)data;
		if ((ret = midi_killable(p, val)) != 0)
			return (ret);
		softc->pgid = val;
		break;
	case TIOCGPGRP:
		*(int *)data = softc->pgid;
		break;
	case MRESET:
		if (!midi_fullreset(softc))
			return (EIO);
		break;
	case MSDIVISION:
		/* must recalculate play remainder */
		val = *(int *)data;
		softc->premainder = softc->premainder * val / softc->division;
		softc->division = val;
		break;
	case MGDIVISION:
		*(int *)data = softc->division;
		break;
	case MDRAIN:
		/* dequeue all everything */
		while (midi_deq(softc->rqueue, &event))
			stynamic_release(&event->data);
		while (midi_deq(softc->wqueue, &event))
			stynamic_release(&event->data);
		/* remove any events already being timed */
		untimeout(midi_timeout, softc);
		softc->status &= ~MIDI_WR_BLOCK;
		softc->status |= MIDI_RD_BLOCK;
		if (softc->status & MIDI_WR_SLEEP) {
			softc->status &= ~MIDI_WR_SLEEP;
			softc->status |= MIDI_WR_ABORT;
			wakeup(softc->wqueue);
		}
		midi_reset_devices(softc);
		break;
	case MFLUSH:
		if (softc->wqueue->count != 0) {
			softc->status |= MIDI_FLUSH_SLEEP;
			do {
				if (error = tsleep(&softc->status,
				    PWAIT | PCATCH, "midiflush", 0))
					return (error);
			} while (softc->status & MIDI_FLUSH_SLEEP);
		}
		break;
	case MGPLAYQ:
		*(int *)data = softc->wqueue->count;
		break;
	case MGRECQ:
		*(int *)data = softc->rqueue->count;
		break;
	case MGQAVAIL:
		*(int *)data = MIDI_Q_SIZE - softc->wqueue->count;
		break;
	case MTHRU:
		if (*(int *)data)
			softc->status |= MIDI_THRU;
		else
			softc->status &= ~MIDI_THRU;
		break;
	case MRECONPLAY:
		if (*(int *)data)
			softc->status |= MIDI_RECONPLAY;
		else
			softc->status &= ~MIDI_RECONPLAY;
		break;
	default:
		return (ENOTTY);
	}
	return (0);
}

int
midiselect(dev, which, p)
	dev_t dev;
	int which;
	struct proc *p;
{
	register struct midi_softc *softc;
	int ret, s, unit;

	unit = minor(dev);
#ifdef __bsdi__
	softc = midicd.cd_devs[unit];
#else
	softc = &midi_sc[unit];
#endif

	s = spltty();
	switch (which) {
	case FREAD:
		if (!(softc->status & MIDI_RD_BLOCK))
			ret = 1;
		else {
			ret = 0;
			selrecord(p, &softc->rsel);
		}
		break;
	case FWRITE:
		if (!(softc->status & MIDI_WR_BLOCK))
			ret = 1;
		else {
			ret = 0;
			selrecord(p, &softc->wsel);
		}
		break;
	default:
		return (0);
	}
	splx(s);
	return (ret);
}

int
midi_fullreset(softc)
	struct midi_softc *softc;
{
	u_char pitch;
	struct event *event;

	/* dequeue all everything */
	while (midi_deq(softc->rqueue, &event))
		stynamic_release(&event->data);
	while (midi_deq(softc->wqueue, &event))
		stynamic_release(&event->data);
	/* remove any events already being timed */
	untimeout(midi_timeout, softc);
	/* mark start time */
	softc->prev_incoming.tv_sec = time.tv_sec;
	softc->prev_incoming.tv_usec = time.tv_usec;
	/* reset fractional count */
	softc->premainder = 0;
	softc->rremainder = 0;
	/* initialize some variables */
	/* this is 120 bpm when a quarter note == 1 beat */
	softc->ptempo = 500000;
	softc->rtempo = 500000;
	softc->prev_rtempo = 500000;

	/* clear noteon */
	for (pitch = 0; pitch <= 0x7f; pitch++)
		softc->noteon[pitch] = 0;
	softc->noteonrs = 0;

	/* clear running play state */
	softc->writers = 0;

	/* reset partial event stuff */
	stynamic_release(&softc->partial_event.event);
	softc->partial_event.state = START;
	stynamic_release(&softc->rpartial);
	stynamic_release(&softc->wpartial);
	softc->wpartialpos = 0;

	/* defaults to 120 clocks per beat */
	softc->division = 120;
	/* reset and enter uart mode */
	return (midi_uart(softc));
}

int
midi_wait_rdy_rcv(softc)
	struct midi_softc *softc;
{
	int flag;
	register int i;

	for (i = 0; i < MIDI_TRIES; i++) {
		flag = INB(softc->addr + MIDI_STATUS) & MIDI_RDY_RCV;
		if (flag == 0)
			break;
	}
	if (i == MIDI_TRIES)
		return (0);
	return(1);
}

int
midi_send_command(softc, comm)
	struct midi_softc *softc;
	u_char comm;
{
	int flag;
	register int i, s;
	unsigned char ack;

	/* writing command and setting ack flag must be atomic */
	s = spltty();
	OUTB(softc->addr + MIDI_COMMAND, comm);
	softc->status |= MIDI_NEED_ACK;
	splx(s);

	/* time out after MIDI_TRIES times */
	for (i = 0; i < MIDI_TRIES; i++) {
		/* did we pick up the ack via midiintr? */
		if (!(softc->status & MIDI_NEED_ACK))
			break;
		/* can we read a data byte? */
		flag = INB(softc->addr + MIDI_STATUS) & MIDI_DATA_AVL;
		if (flag == 0)
			break;
	}
	if (i == MIDI_TRIES)
		return (0);

	if (softc->status & MIDI_NEED_ACK) {
		ack = INB(softc->addr + MIDI_DATA);
		if (ack != MIDI_ACK)
			return (0);
	}
	softc->status &= ~MIDI_NEED_ACK;
	return (1);
}

int
midi_reset(softc)
	struct midi_softc *softc;
{

	if (!midi_wait_rdy_rcv(softc))
		return (0);

	if (!midi_send_command(softc, MIDI_RESET))
		return (0);

	return (1);
}

int
midi_uart(softc)
	struct midi_softc *softc;
{

	/*
	 * first reset.  We can't issue the UART command when already
	 * in UART mode.
	 */
	(void)midi_reset(softc);

	if (!midi_wait_rdy_rcv(softc))
		return (0);

	if (!midi_send_command(softc, MIDI_UART))
		return (0);
	return (1);
}

int
midi_event2smf(softc, event, smf)
	struct midi_softc *softc;
	struct event *event;
	struct stynamic *smf;
{
	long tempo, smf_ticks;
	int tmp_len;
	u_char tmp_buf[4];

	/* convert from kernel ticks to SMF ticks */
	smf_ticks = midi_kernel2smf_tick(softc, event->tempo, event->time);
	tmp_len = midi_fix2var(smf_ticks, tmp_buf);
	stynamic_add_bytes(smf, tmp_buf, tmp_len);

	switch (event->type) {
	case NORMAL:
		stynamic_append(smf, &event->data);
		break;
	case TEMPO:
		/* this case just won't occur */
		stynamic_copy(&event->data, (void *)&tempo, sizeof(tempo));
		stynamic_add_byte(smf, 0xff);
		stynamic_add_byte(smf, 0x51);
		stynamic_add_byte(smf, 0x03);
		stynamic_add_byte(smf, (tempo & 0xff0000) >> 16);
		stynamic_add_byte(smf, (tempo & 0xff00) >> 8);
		stynamic_add_byte(smf, tempo & 0xff);
		break;
	case SYSX:
		stynamic_add_byte(smf, 0xf0);
		/* skip over the leading 0xf0 */
		stynamic_shift(&event->data, 1);
		tmp_len = midi_fix2var(event->data.len, tmp_buf);
		stynamic_add_bytes(smf, tmp_buf, tmp_len);
		stynamic_append(smf, &event->data);
		break;
	}
	return (1);
}

int
midi_next_byte(softc, uio)
	struct midi_softc *softc;
	struct uio *uio;
{
	int byte;

	/* if we're not at the end of a partial event, read from it */
	if (softc->wpartialpos < softc->wpartial.len) {
		byte = stynamic_get_byte(&softc->wpartial,
		    softc->wpartialpos++);
		return (byte);
	} else {
		/* read from uio and copy uio onto partial event */
		if ((byte = uwritec(uio)) == -1) {
			/*
			 * reset partialpos so next time through
			 * we'll read from the partial event if
			 * it is non-zero in length
			 */
			softc->wpartialpos = 0;
			return (-1);
		}
		stynamic_add_byte(&softc->wpartial, byte);
		softc->wpartialpos = softc->wpartial.len;
		return (byte);
	}
}

int
midi_uio2event(softc, uio, event)
	struct midi_softc *softc;
	struct uio *uio;
	struct event *event;
{
	u_long smf_ticks, ulen;
	long tempo;
	int byte, error, extra_byte, i, len, num_data_bytes;
	int rs_change;
	u_char meta_type, tmp_buf[256];

	/* copy in timing portion */
	len = 0;
	do {
		if ((byte = midi_next_byte(softc, uio)) == -1) {
			stynamic_release(&event->data);
			return (-1);
		}
		tmp_buf[len++] = byte;
	} while (byte & 0x80);

	/* compute time in smf ticks */
	midi_var2fix(tmp_buf, &smf_ticks);
	smf_ticks += softc->write_noop_time;
	softc->write_noop_time = 0;

	/* now convert from smf to kernel */
	event->time = midi_smf2kernel_tick(softc, smf_ticks);

	/* get first byte of event data */
	if ((byte = midi_next_byte(softc, uio)) == -1) {
		stynamic_release(&event->data);
		return (-1);
	}
	switch (byte) {
	case 0xf0:
		/* basic sysex type */
		event->type = SYSX;
		len = 0;
		do {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[len++] = byte;
		} while (byte & 0x80);
		midi_var2fix(tmp_buf, &ulen);
		stynamic_add_byte(&event->data, 0xf0);
		for (; ulen > 0; ulen--) {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			stynamic_add_byte(&event->data, byte);
		}
		break;
	case 0xf7:
		/* continued sysex type */
		event->type = SYSX;
		len = 0;
		do {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[len++] = byte;
		} while (byte & 0x80);
		midi_var2fix(tmp_buf, &ulen);
		for (; ulen > 0; ulen--) {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			stynamic_add_byte(&event->data, byte);
		}
		break;
	case 0xff:
		/* meta events */
		if ((byte = midi_next_byte(softc, uio)) == -1) {
			stynamic_release(&event->data);
			return (-1);
		}
		meta_type = byte;
		/* get length of meta data */
		len = 0;
		do {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[len++] = byte;
		} while (byte & 0x80);
		midi_var2fix(tmp_buf, &ulen);
		/* read it in  - meta events are not over 256 in size */
		for (i = 0; i < ulen; i++) {
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[i] = byte;
		}
		switch (meta_type) {
		default:
			/*
			 * we'll skip these events, but we need to
			 * save the timing info
			 */
			softc->write_noop_time = smf_ticks;
			stynamic_release(&softc->wpartial);
			softc->wpartialpos = 0;
			return (-2);
		case 0x51:
			/* tempo event */
			event->type = TEMPO;
			tempo = 0;
			for (i = 0; i < 3; i++) {
				tempo = tempo << 8;
				tempo |= tmp_buf[i];
			}
			stynamic_add_bytes(&event->data, (u_char *)&tempo,
			    sizeof(tempo));
			/*
			 * change ptempo now, rtempo will change when
			 * the event's time comes up
			 */
			softc->ptempo = tempo;
			break;
		}
		break;
	default:
		if ((byte & 0xf0) == 0x80) {
			u_char rs_chan, rs_type;

			/* check note off events separately */
			tmp_buf[0] = byte;
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[1] = byte;
			if ((byte = midi_next_byte(softc, uio)) == -1) {
				stynamic_release(&event->data);
				return (-1);
			}
			tmp_buf[2] = byte;
			len = 3;
			/*
			 * check to see if we can collapse and use
			 * running state
			 */
			rs_type = softc->writers & 0xf0;
			rs_chan = softc->writers & 0x0f;
			/*
			 * collapse to running state if time is 0
			 * and the running state is the same
			 * or the running state is note on for the
			 * same channel and the note off velocity is
			 * zero.
			 */
			if (event->time == 0 && (softc->writers == tmp_buf[0]
			    || (tmp_buf[2] == 0 && rs_type == 0x90
			    && rs_chan == (tmp_buf[0] & 0x0f)))) {
				tmp_buf[0] = tmp_buf[1];
				tmp_buf[1] = tmp_buf[2];
				len = 2;
			} else {
				softc->writers = tmp_buf[0];
			}
		} else {
			extra_byte = 0;
			rs_change = 0;
			if ((byte & 0x80) && (byte != softc->writers)) {
				softc->writers = byte;
				rs_change = 1;
			}
			len = 0;
			if (event->time != 0 || rs_change) {
				/*
				 * stick in a mode byte if time is non-zero
				 * This is so we don't confuse hardware that
				 * is turned on while we're playing
				 * also add it if the running state changes
				 */
				tmp_buf[0] = softc->writers;
				len = 1;
			}
			if (byte & 0x80)
				extra_byte = 1;
			else {
				tmp_buf[len] = byte;
				len++;
			}

			switch (softc->writers & 0xf0) {
			case 0x80:
			case 0x90:
			case 0xa0:
			case 0xb0:
			case 0xe0:
				num_data_bytes = 1;
				break;
			default:
				num_data_bytes = 0;
			}
			for (i = 0; i < num_data_bytes + extra_byte; i++) {
				if ((byte = midi_next_byte(softc, uio)) == -1) {
					stynamic_release(&event->data);
					return (-1);
				}
				tmp_buf[len++] = byte;
			}
		}
		event->type = NORMAL;
		stynamic_add_bytes(&event->data, tmp_buf, len);
	}
	stynamic_release(&softc->wpartial);
	softc->wpartialpos = 0;
	return (0);
}

int
midi_fix2var(fix, var)
	u_long fix;
	u_char *var;
{
	int i;
	unsigned char buf[4], *bptr;

	buf[0] = buf[1] = buf[2] = buf[3] = 0;
	bptr = buf;
	*bptr++ = fix & 0x7f;
	while ((fix >>= 7) > 0) {
		*bptr |= 0x80;
		*bptr++ += (fix & 0x7f);
        }

	i = 0;
	do {
		*var++ = *--bptr;
		i++;
	} while (bptr != buf);

	return (i);
}

int
midi_var2fix(var, fix)
	u_char *var;
	u_long *fix;
{
	int delta;

	*fix = 0;
	delta = 0;
	do {
		*fix = (*fix << 7) + (*var & 0x7f);
		delta++;
	} while (*var++ & 0x80);

	return (delta);
}

u_long
midi_smf2kernel_tick(softc, smf)
	struct midi_softc *softc;
	long smf;
{
	long long denominator, numerator;
	u_long kernel;

	numerator = (long long)softc->ptempo * hz * smf + softc->premainder;
	denominator = 1000000 * (long long)softc->division;
	kernel = numerator / denominator;
	softc->premainder = numerator % denominator;
	return (kernel);
}

u_long
midi_kernel2smf_tick(softc, tempo, kernel)
	struct midi_softc *softc;
	long tempo, kernel;
{
	long long numerator, denominator;
	u_long smf;

	if (softc->prev_rtempo != tempo) {
		/*
		 * also update the rremainder to reflect tempo
		 * change
		 */
		softc->rremainder = softc->rremainder * tempo
		    / softc->prev_rtempo;
		softc->prev_rtempo = tempo;
	}
	numerator = (long long)softc->division * 1000000 * kernel
	    + softc->rremainder;
	denominator = (long long)tempo * hz;
	smf = numerator / denominator;
	softc->rremainder = numerator % denominator;
	return (smf);
}

void
midi_add_complete_event(softc)
	struct midi_softc *softc;
{
	struct timeval t;
	struct timeval curr;
	struct event event;
	struct partial_event *pe;
	long ticks;

	stynamic_init(&event.data);
	pe = &softc->partial_event;
	/* calculate delta time */
	if (softc->prev_incoming.tv_usec > pe->time.tv_usec) {
		curr.tv_usec = pe->time.tv_usec + 1000000;
		curr.tv_sec = pe->time.tv_sec - 1;
	} else {
		curr.tv_usec = pe->time.tv_usec;
		curr.tv_sec = pe->time.tv_sec;
	}
	t.tv_sec = curr.tv_sec - softc->prev_incoming.tv_sec;
	t.tv_usec = curr.tv_usec - softc->prev_incoming.tv_usec;
	if (t.tv_sec >= 0 && t.tv_usec >= 0)
		ticks = t.tv_sec * hz + t.tv_usec / (1000000 / hz);
	else {
		/* real time START commands must have messed things up */
		ticks = 0;
		pe->time.tv_usec = softc->prev_incoming.tv_usec;
		pe->time.tv_sec = softc->prev_incoming.tv_sec;
	}
	event.time = ticks;
	event.tempo = pe->tempo;
	switch (stynamic_get_byte(&pe->event, 0)) {
	case 0xf0:
		/* sysex */
		event.type = SYSX;
		break;
	default:
		event.type = NORMAL;
		break;
	}
	stynamic_append(&event.data, &softc->partial_event.event);
	/* enqueue new event */
	midi_enq(softc->rqueue, &event);
	stynamic_release(&event.data);
	/* readjust previous event time */
	softc->prev_incoming.tv_sec = pe->time.tv_sec;
	softc->prev_incoming.tv_usec = pe->time.tv_usec;

	softc->status &= ~MIDI_RD_BLOCK;
	if (softc->status & MIDI_RD_SLEEP) {
		softc->status &= ~MIDI_RD_SLEEP;
		wakeup(softc->rqueue);
	}
	if (softc->status & MIDI_ASYNC) {
		struct proc *p;

		if (softc->pgid < 0)
			gsignal(-softc->pgid, SIGIO);
		else if ((p = pfind(softc->pgid)) != 0)
			psignal(p, SIGIO);
	}
	selwakeup(&softc->rsel);

	return;
}

void
midi_schedule_timeout(softc)
	struct midi_softc *softc;
{
	struct event *event;

	if (!midi_peekq(softc->wqueue, &event))
		return;
	timeout(midi_timeout, softc, event->time);

	/* reset recording timer? */
	if (softc->status & MIDI_RECONPLAY) {
		softc->prev_incoming.tv_sec = time.tv_sec;
		softc->prev_incoming.tv_usec = time.tv_usec;
		/* clear flag */
		softc->status &= ~MIDI_RECONPLAY;
	}
}

void
midi_timeout(arg)
	caddr_t arg;
{
	struct event *event;
	struct midi_softc *softc = (struct midi_softc *)arg;
	int i;

	/* send first event since we know it is ready */
	midi_deq(softc->wqueue, &event);
	midi_write_event(softc, event);
	stynamic_release(&event->data);

	/* process all events that also occur at this time (event->time == 0) */
	for(;;) {
		if (!midi_peekq(softc->wqueue, &event))
			return;
		if (event->time != 0)
			break;	
		midi_deq(softc->wqueue, NULL);
		midi_write_event(softc, event);
		stynamic_release(&event->data);
	}

	/* setup timeout for next event */
	midi_schedule_timeout(softc);
}

void
midi_write_event(softc, event)
	struct midi_softc *softc;
	struct event *event;
{
	int i, j;
	u_char bytes[4], channel, command, *dataptr;

	switch (event->type) {
	case TEMPO:
		stynamic_copy(&event->data, &softc->rtempo,
		    sizeof(softc->rtempo));
		break;
	case NORMAL:
		/*
		 * fourth byte might not be valid, but who cares,
		 * we're only reading and in the kernel.  We'll
		 * ignore it if it isn't.
		 */
		stynamic_copy(&event->data, &bytes, 4);
		if (!(bytes[0] & 0x80))
			dataptr = &bytes[0];
		else {
			softc->noteonrs = bytes[0];
			dataptr = &bytes[1];
		}
		command = softc->noteonrs & 0xf0;
		channel = softc->noteonrs & 0x0f;
		if (command == 0x90) {
			/*
			 * set of clear appropriate bit in noteon
			 * array depending on velocity value
			 */
			if (dataptr[1] != 0)
				softc->noteon[dataptr[0]] |= 1 << channel;
			else
				softc->noteon[dataptr[0]] &= ~(1 << channel);
		}
		if (command == 0x80)
			/* clear bit */
			softc->noteon[dataptr[0]] &= ~(1 << channel);
		/* FALLTHRU */
	default:
		for (i = 0; i < event->data.len; i++) {
			if (!midi_wait_rdy_rcv(softc))
				break;
			OUTB(softc->addr + MIDI_DATA,
			    stynamic_get_byte(&event->data, i));
		}
		break;
	}
	if (softc->wqueue->count < MIDI_LOW_WATER) {
		softc->status &= ~MIDI_WR_BLOCK;
		if (softc->status & MIDI_WR_SLEEP) {
			softc->status &= ~MIDI_WR_SLEEP;
			wakeup(softc->wqueue);
		}
		if (softc->status & MIDI_ASYNC &&
		    (softc->status & MIDI_SENDIO || softc->wqueue->count
		    == 0)) {
			struct proc *p;

			if (softc->pgid < 0)
				gsignal(-softc->pgid, SIGIO);
			else if ((p = pfind(softc->pgid)) != 0)
				psignal(p, SIGIO);
		}
		softc->status &= ~MIDI_SENDIO;
		selwakeup(&softc->wsel);
	}
	if (softc->status & MIDI_FLUSH_SLEEP && softc->wqueue->count == 0) {
		softc->status &= ~MIDI_FLUSH_SLEEP;
		wakeup(&softc->status);
	}
}


/*
 * try to reset the midi devices as best we can
 */
void
midi_reset_devices(softc)
	struct midi_softc *softc;
{
	int i;
	u_char channel, pitch;

	/* manual note off calls - turn off any note that is on */
	for (pitch = 0; pitch <= 0x7f; pitch++) {
		for (channel = 0; channel <= 0x0f; channel++) {
			if ((softc->noteon[pitch] >> channel) & 0x01) {
				if (!midi_wait_rdy_rcv(softc))
					break;
				OUTB(softc->addr + MIDI_DATA, channel | 0x90);

				if (!midi_wait_rdy_rcv(softc))
					break;
				OUTB(softc->addr + MIDI_DATA, pitch);

				if (!midi_wait_rdy_rcv(softc))
					break;
				OUTB(softc->addr + MIDI_DATA, 0);
			}
		}
		softc->noteon[pitch] = 0;
	}
	for (channel = 0; channel <= 0x0f; channel++) {
		/*
		 * send paramter event for all notes off for redundancy
		 * some older synths don't support this
		 */
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, channel | 0xb0);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0x7b);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0);

		/* modulation controller to zero */
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0x01);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0);

		/* reset all controllers */
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0x79);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0);

		/* lift sustain pedal */
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0x40);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0);

		/* center pitch wheel */
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0xe0 | channel);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0);
		if (!midi_wait_rdy_rcv(softc))
			break;
		OUTB(softc->addr + MIDI_DATA, 0x40);
	}
	softc->noteonrs = 0;
}

void
midi_initq(eq)
	struct event_queue *eq;
{

	eq->count = 0;
	eq->end = &eq->events[MIDI_Q_SIZE - 1];
	eq->head = eq->events;
	eq->tail = eq->events;
	/* zero events to clear stynamic stuctures */
	bzero(eq->events, MIDI_Q_SIZE * sizeof(struct event));
}

void
midi_copy_event(e1, e2)
	struct event *e1, *e2;
{
	short i;

	e1->time = e2->time;
	e1->type = e2->type;
	e1->tempo = e2->tempo;
	stynamic_release(&e1->data);
	stynamic_append(&e1->data, &e2->data);
}

int
midi_deq(eq, event)
	struct event_queue *eq;
	struct event **event;
{
	int s;

	s = spltty();
	if (eq->count == 0) {
		splx(s);
		return (0);
	}
	if (event == NULL)
		eq->head++;
	else
		*event = eq->head++;
	if (eq->head > eq->end)
		eq->head = eq->events;
	eq->count--;
	splx(s);
	return (1);
}

int
midi_peekq(eq, event)
	struct event_queue *eq;
	struct event **event;
{
	int s;

	s = spltty();
	if (eq->count == 0) {
		splx(s);
		return (0);
	}
	*event = eq->head;
	splx(s);
	return (1);
}

int
midi_enq(eq, event)
	struct event_queue *eq;
	struct event *event;
{
	int s;

	s = spltty();
	if (eq->count == MIDI_Q_SIZE) {
		splx(s);
		return (0);
	}
	midi_copy_event(eq->tail++, event);
	if (eq->tail > eq->end)
		eq->tail = eq->events;
	eq->count++;
	splx(s);
	return (1);
}

void
stynamic_add_byte(sty, byte)
	struct stynamic *sty;
	u_char byte;
{
	int s;
	u_char *new_ptr;

	s = spltty();
	if (sty->len < STYNAMIC_SIZE)
		sty->datas[sty->len++] = byte;
	else {
		if (sty->len < sty->allocated) {
			sty->datad[sty->len++] = byte;
		} else {
			new_ptr = malloc(sty->allocated + STYNAMIC_ALLOC,
			    M_DEVBUF, M_WAITOK);
			if (sty->allocated == 0)
				bcopy(sty->datas, new_ptr, sty->len);
			else {
				bcopy(sty->datad, new_ptr, sty->len);
				free(sty->datad, M_DEVBUF);
			}
			sty->datad = new_ptr;
			sty->datad[sty->len++] = byte;
			sty->allocated += STYNAMIC_ALLOC;
		}
	}
	splx(s);
}

void
stynamic_add_bytes(sty, bytes, num)
	struct stynamic *sty;
	u_char *bytes;
	int num;
{
	int s, size_inc;
	u_char *new_ptr;

	s = spltty();
	if (sty->len + num <= STYNAMIC_SIZE) {
		bcopy(bytes, &sty->datas[sty->len], num);
		sty->len += num;
	} else {
		if (sty->len + num <= sty->allocated) {
			bcopy(bytes, &sty->datad[sty->len], num);
			sty->len += num;
		} else {
			size_inc = (num / STYNAMIC_ALLOC + 1) * STYNAMIC_ALLOC;
			new_ptr = malloc(sty->allocated + size_inc,
			    M_DEVBUF, M_WAITOK);
			if (sty->allocated == 0)
				bcopy(sty->datas, new_ptr, sty->len);
			else {
				bcopy(sty->datad, new_ptr, sty->len);
				free(sty->datad, M_DEVBUF);
			}
			sty->datad = new_ptr;
			bcopy(bytes, &sty->datad[sty->len], num);
			sty->allocated += size_inc;
			sty->len += num;
		}
	}
	splx(s);
}

u_char
stynamic_get_byte(sty, index)
	struct stynamic *sty;
	int index;
{
	int s;

	s = spltty();
	if (sty->len <= 4)
		return (sty->datas[index]);
	else
		return (sty->datad[index]);
	splx(s);
}

void
stynamic_copy(sty, dest, len)
	struct stynamic *sty;
	void *dest;
	int len;
{
	int s;

	s = spltty();
	if (sty->len <= 4)
		bcopy(sty->datas, dest, len);
	else
		bcopy(sty->datad, dest, len);
	splx(s);
}

void
stynamic_append(dest, src)
	struct stynamic *dest;
	struct stynamic *src;
{

	if (src->len <= 4)
		stynamic_add_bytes(dest, src->datas, src->len);
	else
		stynamic_add_bytes(dest, src->datad, src->len);
}

void
stynamic_copy_from(sty, index, dest, len)
	struct stynamic *sty;
	int index, len;
	void *dest;
{
	int s;

	s = spltty();
	if (sty->len <= 4)
		bcopy(&sty->datas[index], dest, len);
	else
		bcopy(&sty->datad[index], dest, len);
	splx(s);
}


void
stynamic_release(sty)
	struct stynamic *sty;
{
	int s;

	s = spltty();
	if (sty->len > STYNAMIC_SIZE)
		free(sty->datad, M_DEVBUF);
	sty->len = 0;
	sty->allocated = 0;
	sty->datad = NULL;
	bzero(sty->datas, STYNAMIC_SIZE);
	splx(s);
}

void
stynamic_shift(sty, num)
	struct stynamic *sty;
	int num;
{
	int rem, s;
	u_char *ptr;

	if (sty->len <= num) {
		stynamic_release(sty);
		return;
	}
	s = spltty();
	if (sty->len > STYNAMIC_SIZE)
		ptr = &sty->datad[num];
	else
		ptr = &sty->datas[num];
	rem = sty->len - num;
	if (rem > STYNAMIC_SIZE)
		bcopy(ptr, sty->datad, rem);
	else {
		bcopy(ptr, sty->datas, rem);
		if (sty->datad != NULL) {
			free(sty->datad, M_DEVBUF);
			sty->datad = NULL;
			sty->allocated = 0;
		}
	}
	sty->len = rem;
	splx(s);
}

void
stynamic_init(sty)
	struct stynamic *sty;
{
	int s;

	s = spltty();
	sty->len = 0;
	sty->allocated = 0;
	sty->datad = NULL;
	bzero(sty->datas, STYNAMIC_SIZE);
	splx(s);
}

void
stynamic_print(sty)
	struct stynamic *sty;
{

	printf("\t\tlen = %d\n", sty->len);
	printf("\t\tallocated = %d\n", sty->allocated);
	printf("\t\tdatad = 0x%x\n", sty->datad);
}

#define CANSIGNAL(p, q) \
	((p)->p_cred->pc_ucred->cr_uid == 0 || \
	    (p)->p_cred->p_ruid == (q)->p_cred->p_ruid || \
	    (p)->p_cred->pc_ucred->cr_uid == (q)->p_cred->p_ruid || \
	    (p)->p_cred->p_ruid == (q)->p_ucred->cr_uid || \
	    (p)->p_cred->pc_ucred->cr_uid == (q)->p_ucred->cr_uid)
int
midi_killable(owner, pid)
	struct proc *owner;
	int pid;
{
	register struct proc *p;
	struct pgrp *pgrp;

	if (pid == 0)
		return (0);

	/* let's not even deal with broadcasts */
	if (pid == -1)
		return (EPERM);
	if (pid < 0) {
		pgrp = pgfind(-pid);
		if (pgrp == NULL)
			return (ESRCH);
		for (p = pgrp->pg_mem; p != NULL; p = p->p_pgrpnxt) {
			if (!CANSIGNAL(owner, p))
				return (EPERM);
		}
		return (0);
	}
	p = pfind(pid);
	if (!CANSIGNAL(owner, p))
		return (EPERM);
	return (0);
}
#endif
